from concurrent.futures import ThreadPoolExecutor
from email import message_from_bytes
from email.header import decode_header, make_header
from imaplib import IMAP4_SSL
from smtplib import SMTP, SMTPAuthenticationError, SMTPConnectError
from threading import Thread

from gmailconnector.responder import Response


class Messenger:
    """Initiates Messenger object to send an SMS to a phone number using SMS gateway provided by the mobile carrier.

    >>> Messenger

    Args:
        gmail_user: Gmail username to authenticate SMTP lib.
        gmail_pass: Gmail password to authenticate SMTP lib.
        phone: Phone number stored as env var.
        message: Content of the message.
        subject [Optional] : Subject line for the message. Defaults to "Message from GmailConnector"
        carrier [Optional]: Takes any of ``at&t``, ``t-mobile``, ``verizon``, ``boost``, ``cricket``, ``us-cellular``
        sms_gateway [Optional]: Takes the SMS gateway of the carrier as an argument.

    See Also:
        Carrier defaults to ``t-mobile`` which uses ``tmomail.net`` as the SMS gateway.
    """

    SMS_GATEWAY = {
        "at&t": "mms.att.net",
        "t-mobile": "tmomail.net",
        "verizon": "vtext.com",
        "boost": "smsmyboostmobile.com",
        "cricket": "sms.cricketwireless.net",
        "us-cellular": "email.uscc.net",
    }

    def __init__(self, gmail_user: str, gmail_pass: str, phone: str, message: str,
                 subject: str = None, carrier: str = 't-mobile', sms_gateway: str = None,
                 delete_sent: bool = True):
        self.username = gmail_user
        self.password = gmail_pass
        self.phone = phone
        self.delete_sent = delete_sent

        self.subject = subject or "Message from GmailConnector"
        self.message = message
        self.server = SMTP("smtp.gmail.com", 587)

        self._arg_validator()

        self.body = f'\n\n{message}'.encode('ascii', 'ignore').decode('ascii')

        carrier = carrier.lower()
        if carrier not in list(self.SMS_GATEWAY.keys()):
            carrier = 't-mobile'
        self.gateway = sms_gateway or self.SMS_GATEWAY.get(carrier)

        self.to = self._generate_address()
        self.mail = None

    def __del__(self):
        """Destructor has been called to close the TLS connection and logout."""
        if self.server:
            self.server.close()

    def _arg_validator(self):
        """Validates all the arguments passed during object initialization.

        Raises:
            ValueError: If any arg is missing or malformed.
        """
        if not all([self.username, self.password, self.phone, self.message]):
            raise ValueError(
                'Cannot proceed without the args: `gmail_user`, `gmail_pass`, `phone` and `message`'
            )
        if len(self.phone) != 10 and len(self.phone) != 12:
            raise ValueError('Phone number should either be 10 or 12 digits (if includes country code)')
        if self.phone.startswith('+') and not self.phone.startswith('+1'):
            raise ValueError('Unsupported country code. Module works only for US based contact numbers.')

    def _generate_address(self) -> str:
        """Validates the digits in the phone number and forms the endpoint using the phone number and sms gateway.

        Returns:
            str:
            Returns the formed endpoint. `Example: ``+11234567890@tmomail.net```
        """
        if len(self.phone) == 11 and self.phone.startswith('1'):
            self.phone = f'+{self.phone}'
        if not self.phone.startswith('+'):
            self.phone = f'+1{self.phone}'
        return self.phone + '@' + self.gateway

    def send_sms(self) -> Response:
        """Initiates a TLS connection and sends a text message through SMS gateway of destination number.

        Raises:
            UnicodeEncodeError

        See Also:
            - Encodes body of the message to `ascii` with `ignore` flag and then decodes it.
            - This is done to ignore special characters (like °) without raising `UnicodeEncodeError`
            - Validates the endpoint ``phone-number@sms-gateway`` before trying to send the SMS.

        Notes:
            Other flags that can be set includes `replace` and `xmlcharrefreplace`

        Returns:
            Response:
            A custom response class with properties: ok, status and body to the user.
        """
        message = (f"From: {self.username}\n" + f"To: {self.to}\n" + f"Subject: {self.subject}\n" + self.body)

        if len(message) > 428:
            return Response(dictionary={
                'ok': False,
                'status': 413,
                'body': f'Payload length: {len(message):,}, which is more than the optimal size: 428. '
                        f'Message length: {len(self.body):,}'
            })

        self.server.starttls()

        try:
            self.server.login(user=self.username, password=self.password)
        except SMTPAuthenticationError:
            self.server = None
            return Response(dictionary={
                'ok': False,
                'status': 401,
                'body': "GMAIL login failed with SMTPAuthenticationError: Username and Password not accepted.\n"
                        "Ensure the credentials stored in env vars are set correct.\n"
            })
        except SMTPConnectError:
            self.server = None
            return Response(dictionary={
                'ok': False,
                'status': 503,
                'body': "Error during connection establishment with GMAIL server."
            })

        self.server.sendmail(self.username, self.to, message)

        if self.delete_sent:
            Thread(target=self._delete_sent, daemon=True).start()

        return Response(dictionary={
            'ok': True,
            'status': 200,
            'body': f'SMS has been sent to {self.to}'
        })

    def _thread_executor(self, item_id):
        """Gets invoked as multi-threading, to check for the message which was just sent and sets the flag as Deleted.

        Args:
            item_id: Takes the ID of the message as an argument.
        """
        dummy, data = self.mail.fetch(item_id, '(RFC822)')
        for response_part in data:
            if not isinstance(response_part, tuple):
                continue
            original_email = message_from_bytes(response_part[1])  # gets the raw content
            sender = str(make_header(decode_header((original_email['From']).split(' <')[0])))
            sub = str(make_header(decode_header(original_email['Subject'])))
            to = str(make_header(decode_header(original_email['To'])))
            if to == self.to and sub == self.subject and sender == self.username:
                self.mail.store(item_id.decode('UTF-8'), '+FLAGS', '\\Deleted')
                self.mail.expunge()
                self.mail.close()
                self.mail.logout()
                break

    def _delete_sent(self):
        """Deletes the email from GMAIL's sent items right after sending the message.

        See Also:
            Invokes ``thread_executor`` with 20 workers to sweep through sent items to delete the email.
        """
        self.mail = IMAP4_SSL('imap.gmail.com')
        self.mail.login(user=self.username, password=self.password)
        self.mail.list()
        self.mail.select('"[Gmail]/Sent Mail"')
        return_code, messages = self.mail.search(None, 'ALL')  # Includes SEEN and UNSEEN, although sent is always SEEN
        with ThreadPoolExecutor(max_workers=20) as executor:
            executor.map(self._thread_executor, sorted(messages[0].split(), reverse=True))


if __name__ == '__main__':
    from datetime import datetime
    from os import environ

    response = Messenger(
        gmail_user=environ.get('gmail_user'), gmail_pass=environ.get('gmail_pass'),
        phone=environ.get('phone'), message=f'Hello on {datetime.now().strftime("%B %d, %Y at %I:%M %p")}'
    ).send_sms()

    if response.ok and response.status == 200:
        print('SUCCESS')
    else:
        print('FAILED')
    print(response.json())



from datetime import datetime, timedelta
from email import message_from_bytes, message_from_string
from email.header import decode_header, make_header
from imaplib import IMAP4_SSL

from gmailconnector.responder import Response


class ReadEmail:
    """Initiates Emailer object to authenticate and print the unread emails.

    >>> ReadEmail

    Args:
        gmail_user: Email address (GMAIL)
        gmail_pass: Login password
    """

    def __init__(self, gmail_user: str, gmail_pass: str):
        """Gathers all the necessary parameters to read emails."""
        self.mail = IMAP4_SSL('imap.gmail.com')  # connects to gmail using imaplib
        # noinspection PyBroadException
        try:
            self.mail.login(user=gmail_user, password=gmail_pass)  # login to your gmail account using the env variables
            self.mail.list()  # list all the folders within your mailbox (like inbox, sent, drafts, etc)
            self.mail.select('inbox')  # selects inbox from the list
        except Exception:
            self.mail = None
        self.username = gmail_user

    def __del__(self):
        """Destructor called to close the mailbox and logout."""
        if self.mail:
            self.mail.close()
            self.mail.logout()

    def _instantiate(self) -> Response:
        """Prints the number of emails and gets user confirmation before proceeding, press N/n to quit.

        Returns:
            Response:
            A tuple containing number of email messages, return code and the messages itself.
        """
        return_code, messages = self.mail.search(None, 'UNSEEN')  # looks for unread emails
        n = len(messages[0].split())

        if return_code != 'OK':
            return Response(dictionary={
                'ok': False,
                'status': 404,
                'body': 'Unable to read messages.'
            })

        if not n:
            return Response(dictionary={
                'ok': False,
                'status': 204,
                'body': f'You have no unread emails. Account username: {self.username}'
            })

        if return_code == 'OK':
            user_ip = input(f'You have {n} unread emails. Press Y/y to continue:\n')
            if user_ip.upper() == 'Y':
                return Response(dictionary={
                    'ok': True,
                    'status': 200,
                    'body': messages
                })
            else:
                return Response(dictionary={
                    'ok': False,
                    'status': 202,
                    'body': 'User declined to read the email.'
                })
        else:
            return Response(dictionary={
                'ok': False,
                'status': 401,
                'body': "Unable access your email account."
            })

    def read_email(self) -> Response:
        """Prints unread emails one by one after getting user confirmation.

        Returns:
            Response:
            A custom response class with properties: ok, status and body to the user.
        """
        if not self.mail:
            return Response(dictionary={
                'ok': False,
                'status': 401,
                'body': 'BUMMER! Unable to read your emails.\n\nTroubleshooting Steps:\n'
                        '\t1. Make sure your username and password are correct.\n'
                        '\t2. Logon to https://myaccount.google.com/lesssecureapps and turn ON less secure apps.\n'
                        '\t3. If you have enabled 2 factor authentication, use thee App Password generated.'
            })

        initiator = self._instantiate()
        if initiator.ok:
            messages = initiator.body
        else:
            return initiator

        n = len(messages[0].split())
        i = 0
        for nm in messages[0].split():
            i += 1
            dummy, data = self.mail.fetch(nm, '(RFC822)')
            for response_part in data:
                if isinstance(response_part, tuple):
                    original_email = message_from_bytes(response_part[1])  # gets the rawest content
                    date = (original_email['Received'].split(';')[-1]).strip()  # gets raw received time
                    if '(PDT)' in date:
                        datetime_obj = datetime.strptime(date, "%a, %d %b %Y %H:%M:%S -0700 (PDT)") + timedelta(hours=2)
                    elif '(PST)' in date:
                        datetime_obj = datetime.strptime(date, "%a, %d %b %Y %H:%M:%S -0800 (PST)") + timedelta(hours=2)
                    else:
                        datetime_obj = datetime.now()
                    receive = (datetime_obj.strftime("on %A, %B %d, %Y at %I:%M %p CT"))  # formats datetime
                    # noinspection PyUnresolvedReferences
                    raw_email = data[0][1]
                    raw_email_string = raw_email.decode('utf-8')  # decodes the raw email
                    email_message = message_from_string(raw_email_string)  # extracts message from string
                    # generates the file names in a directory tree by walking the tree bottom-up
                    # it yields a 3-tuple (body, sender, subject).
                    for part in email_message.walk():
                        """The 'if' condition below is important as parts of an email may include
                        text/html, multipart/alternative, text/plain and other attachments.
                        We choose only text/plain as it is the only code-friendly part of a multipart email.
                        Choosing text/html can make the output look messed up with unnecessary html tags
                        Choosing multipart/alternative leaves us with decoding errors.
                        Though I have an exception handler to handle the decoding exceptions,
                        if the condition below is removed the loop will keep running through all the parts
                        within a single email"""
                        if part.get_content_type() == "text/plain":  # ignore attachments/html/multipart
                            # returns message's entire payload, a string or a message instance.
                            # decode=true decodes the payload received.
                            body = part.get_payload(decode=True)
                            # gets sender email and subject and removes unnecessary white spaces
                            sender = make_header(decode_header((original_email['From']).split(' <')[0]))
                            sub = make_header(decode_header(original_email['Subject'])) \
                                if original_email['Subject'] else None
                            print(f"You have an email from {sender} with subject '{sub}' {receive}")
                            #  gets user confirmation before printing the decoded body of the email
                            try:
                                msg = (body.decode('utf-8')).strip()  # decodes body of the email
                                get_user_input = input('Enter Y/N to read the email:\n')
                                if get_user_input == 'Y' or get_user_input == 'y':
                                    print(f'{msg}\n')
                            except (UnicodeDecodeError, AttributeError):  # catches both the decoding errors
                                continue
                            if i < n:  # proceeds only if loop count is less than the number of unread emails
                                continue_confirmation = input('Enter N/n to quit, any other key to continue:\n')
                                if continue_confirmation == 'N' or continue_confirmation == 'n':
                                    return Response(dictionary={
                                        'ok': False,
                                        'status': 202,
                                        'body': 'User declined to continue reading emails.'
                                    })
                            else:
                                return Response(dictionary={
                                    'ok': True,
                                    'status': 200,
                                    'body': 'All messages have been read.'
                                })


if __name__ == '__main__':
    from os import environ

    response = ReadEmail(gmail_user=environ.get('gmail_user'), gmail_pass=environ.get('gmail_pass')).read_email()
    print(response.json())



from email.mime import multipart, text
from email.mime.application import MIMEApplication
from os.path import isfile, realpath
from smtplib import SMTP, SMTPAuthenticationError, SMTPConnectError

from gmailconnector.responder import Response


class SendEmail:
    """Initiates Emailer object to send an email to defined recipient from a defined sender with or without attachments.

    >>> SendEmail

    Args:
        gmail_user: Username to login to TLS.
        gmail_pass: Password to authenticate TLS session.
        recipient: Email address of the recipient to whom the email has to be sent.
        subject: Subject line of the email.
        body [Optional] : Body of the email. Defaults to no body.
        attachment [Optional] : Filename that has to be attached.
        cc [Optional]: Email address of the recipient to whom the email has to be CC'd.
        bcc [Optional]: Email address of the recipient to whom the email has to be BCC'd.
    """

    def __init__(self, gmail_user: str, gmail_pass: str, recipient: str or list,
                 subject: str, body: str = None, attachment: str = None,
                 cc: str or list = None, bcc: str or list = None):
        """Gathers all the necessary parameters to send an email."""
        self.gmail_user = gmail_user
        self.gmail_pass = gmail_pass
        self.recipient = recipient
        self.subject = subject
        self.cc = cc
        self.bcc = bcc
        self.body = body
        self.sender = f"GmailConnector <{gmail_user}>"
        self.attachment = attachment
        self.file_not_available = None
        self.server = SMTP(host='smtp.gmail.com', port=587)

    def __del__(self):
        """Destructor has been called to close the TLS connection and logout."""
        if self.server:
            self.server.close()

    def _multipart_message(self) -> multipart.MIMEMultipart:
        """Creates a multipart message with subject, body, from and to address, and attachment if filename is passed.

        Returns a message if a filename is given for attachment but not available at the given path.

        Returns:
            `multipart.MIMEMultipart`:
            MIMEMultipart version of the created message.

        """
        msg = multipart.MIMEMultipart()
        msg['Subject'] = self.subject
        msg['From'] = self.sender
        msg['To'] = ','.join(self.recipient) if isinstance(self.recipient, list) else self.recipient
        if cc := self.cc:
            msg['Cc'] = cc

        if body := self.body:
            msg.attach(payload=text.MIMEText(body))

        if (filename := self.attachment) and (isfile(filename)):
            with open(filename, 'rb') as attachment:
                attribute = MIMEApplication(attachment.read(), _subtype=filename.split('.')[-1])
            attribute.add_header('Content-Disposition', 'attachment', filename=filename)
            msg.attach(payload=attribute)
        elif filename:
            self.file_not_available = True

        return msg

    def send_email(self) -> Response:
        """Initiates a TLS connection and sends the email.

        Returns:
            Response:
            A custom response class with properties: ok, status and body to the user.
        """
        self.server.starttls()
        try:
            self.server.login(user=self.gmail_user, password=self.gmail_pass)
        except SMTPAuthenticationError:
            self.server = None
            return Response(dictionary={
                'ok': False,
                'status': 401,
                'body': "GMAIL login failed with SMTPAuthenticationError: Username and Password not accepted.\n"
                        "Ensure the credentials stored in env vars are set correct.\n"
            })
        except SMTPConnectError:
            self.server = None
            return Response(dictionary={
                'ok': False,
                'status': 503,
                'body': "Error during connection establishment with GMAIL server."
            })

        to = self.recipient
        recipients = [to] if isinstance(to, str) else to

        if cc := self.cc:
            recipients.append(cc) if isinstance(cc, str) else recipients.extend(cc)

        if bcc := self.bcc:
            recipients.append(bcc) if isinstance(bcc, str) else recipients.extend(bcc)

        self.server.sendmail(
            from_addr=self.sender,
            to_addrs=recipients,
            msg=self._multipart_message().as_string()
        )

        return_msg = f'Email has been sent to {self.recipient}'

        if self.file_not_available:
            return Response(dictionary={
                'ok': True,
                'status': 206,
                'body': return_msg + f"\n{self.attachment} is unavailable at {realpath(filename='')}.\n"
                                     f"Email was sent without an attachment."
            })
        else:
            return Response(dictionary={
                'ok': True,
                'status': 200,
                'body': return_msg
            })


if __name__ == '__main__':
    from datetime import datetime
    from os import environ

    response = SendEmail(
        gmail_user=environ.get('gmail_user'), gmail_pass=environ.get('gmail_pass'), recipient=environ.get('recipient'),
        subject=datetime.now().strftime("%B %d, %Y %I:%M %p")
    ).send_email()

    if response.ok and response.status == 200:
        print('SUCCESS')
    elif response.status == 403:
        print('AUTH ERROR')
    elif response.status == 503:
        print('SERVICE UNAVAILABLE')
    print(response.json())
